"""Composer models
"""
from django.db import models

from core_composer_app.permissions import rights
from core_main_app.permissions.utils import get_formatted_name


class Composer(models.Model):
    class Meta(object):
        verbose_name = "core_composer_app"
        default_permissions = ()
        permissions = (
            (rights.composer_access, get_formatted_name(rights.composer_access)),
            (
                rights.composer_save_template,
                get_formatted_name(rights.composer_save_template),
            ),
            (rights.composer_save_type, get_formatted_name(rights.composer_save_type)),
        )



""" Authentication tests for Type Version Manager REST API
"""
from django.test import SimpleTestCase
from mock.mock import patch
from rest_framework import status

from core_composer_app.components.type_version_manager.models import TypeVersionManager
from core_composer_app.rest.type_version_manager import (
    views as type_version_manager_views,
)
from core_composer_app.rest.type_version_manager.serializers import (
    TypeVersionManagerSerializer,
    CreateTypeSerializer,
)
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from core_main_app.utils.tests_tools.RequestMock import RequestMock


class TestGlobalTypeVersionManagerListGetPermission(SimpleTestCase):
    @patch.object(TypeVersionManager, "get_global_version_managers")
    def test_anonymous_returns_http_403(
        self, type_version_manager_get_all_global_version_managers
    ):
        type_version_manager_get_all_global_version_managers.return_value = {}

        response = RequestMock.do_request_get(
            type_version_manager_views.GlobalTypeVersionManagerList.as_view(), None
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch.object(TypeVersionManager, "get_global_version_managers")
    def test_authenticated_returns_http_200(
        self, type_version_manager_get_all_global_version_managers
    ):
        type_version_manager_get_all_global_version_managers.return_value = {}

        mock_user = create_mock_user("1")

        response = RequestMock.do_request_get(
            type_version_manager_views.GlobalTypeVersionManagerList.as_view(), mock_user
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(TypeVersionManager, "get_global_version_managers")
    def test_staff_returns_http_200(
        self, type_version_manager_get_all_global_version_managers
    ):
        type_version_manager_get_all_global_version_managers.return_value = {}

        mock_user = create_mock_user("1", is_staff=True)

        response = RequestMock.do_request_get(
            type_version_manager_views.GlobalTypeVersionManagerList.as_view(), mock_user
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)


class TestUserTypeVersionManagerListGetPermission(SimpleTestCase):
    def test_anonymous_returns_http_403(self):
        response = RequestMock.do_request_get(
            type_version_manager_views.UserTypeVersionManagerList.as_view(), None
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch.object(TypeVersionManager, "get_version_managers_by_user")
    def test_authenticated_returns_http_200(
        self, type_version_manager_get_version_managers_by_user
    ):
        type_version_manager_get_version_managers_by_user.return_value = {}

        mock_user = create_mock_user("1")

        response = RequestMock.do_request_get(
            type_version_manager_views.UserTypeVersionManagerList.as_view(), mock_user
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(TypeVersionManager, "get_version_managers_by_user")
    def test_staff_returns_http_200(
        self, type_version_manager_get_version_managers_by_user
    ):
        type_version_manager_get_version_managers_by_user.return_value = {}

        mock_user = create_mock_user("1", is_staff=True)

        response = RequestMock.do_request_get(
            type_version_manager_views.UserTypeVersionManagerList.as_view(), mock_user
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)


class TestUserTypeListPostPermission(SimpleTestCase):
    def test_anonymous_returns_http_403(self):
        response = RequestMock.do_request_post(
            type_version_manager_views.UserTypeList.as_view(), None
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch.object(CreateTypeSerializer, "is_valid")
    @patch.object(CreateTypeSerializer, "save")
    @patch.object(CreateTypeSerializer, "data")
    @patch.object(TypeVersionManagerSerializer, "is_valid")
    @patch.object(TypeVersionManagerSerializer, "save")
    @patch.object(TypeVersionManagerSerializer, "data")
    def test_authenticated_returns_http_201(
        self,
        type_version_manager_serializer_data,
        type_version_manager_serializer_save,
        type_version_manager_serializer_valid,
        create_type_version_manager_serializer_data,
        create_type_version_manager_serializer_save,
        create_type_version_manager_serializer_valid,
    ):
        create_type_version_manager_serializer_data.return_value = True
        create_type_version_manager_serializer_save.return_value = None
        create_type_version_manager_serializer_valid.return_value = {}
        type_version_manager_serializer_data.return_value = True
        type_version_manager_serializer_save.return_value = None
        type_version_manager_serializer_valid.return_value = {}

        mock_user = create_mock_user("1")

        response = RequestMock.do_request_post(
            type_version_manager_views.UserTypeList.as_view(), mock_user, data={}
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    @patch.object(CreateTypeSerializer, "is_valid")
    @patch.object(CreateTypeSerializer, "save")
    @patch.object(CreateTypeSerializer, "data")
    @patch.object(TypeVersionManagerSerializer, "is_valid")
    @patch.object(TypeVersionManagerSerializer, "save")
    @patch.object(TypeVersionManagerSerializer, "data")
    def test_staff_returns_http_201(
        self,
        type_version_manager_serializer_data,
        type_version_manager_serializer_save,
        type_version_manager_serializer_valid,
        create_type_version_manager_serializer_data,
        create_type_version_manager_serializer_save,
        create_type_version_manager_serializer_valid,
    ):
        create_type_version_manager_serializer_data.return_value = True
        create_type_version_manager_serializer_save.return_value = None
        create_type_version_manager_serializer_valid.return_value = {}
        type_version_manager_serializer_data.return_value = True
        type_version_manager_serializer_save.return_value = None
        type_version_manager_serializer_valid.return_value = {}

        mock_user = create_mock_user("1", is_staff=True)

        response = RequestMock.do_request_post(
            type_version_manager_views.UserTypeList.as_view(), mock_user, data={}
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)


class TestGlobalTypeListPostPermission(SimpleTestCase):
    def test_anonymous_returns_http_403(self):
        response = RequestMock.do_request_post(
            type_version_manager_views.GlobalTypeList.as_view(), None
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_authenticated_returns_http_403(self):
        mock_user = create_mock_user("1")

        response = RequestMock.do_request_post(
            type_version_manager_views.GlobalTypeList.as_view(), mock_user, data={}
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch.object(CreateTypeSerializer, "is_valid")
    @patch.object(CreateTypeSerializer, "save")
    @patch.object(CreateTypeSerializer, "data")
    @patch.object(TypeVersionManagerSerializer, "is_valid")
    @patch.object(TypeVersionManagerSerializer, "save")
    @patch.object(TypeVersionManagerSerializer, "data")
    def test_staff_returns_http_201(
        self,
        type_version_manager_serializer_data,
        type_version_manager_serializer_save,
        type_version_manager_serializer_valid,
        create_type_version_manager_serializer_data,
        create_type_version_manager_serializer_save,
        create_type_version_manager_serializer_valid,
    ):
        create_type_version_manager_serializer_data.return_value = True
        create_type_version_manager_serializer_save.return_value = None
        create_type_version_manager_serializer_valid.return_value = {}
        type_version_manager_serializer_data.return_value = True
        type_version_manager_serializer_save.return_value = None
        type_version_manager_serializer_valid.return_value = {}

        mock_user = create_mock_user("1", is_staff=True)

        response = RequestMock.do_request_post(
            type_version_manager_views.GlobalTypeList.as_view(), mock_user, data={}
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)



"""Composer AJAX admin views
"""
import json

from django.contrib.admin.views.decorators import staff_member_required
from django.http.response import HttpResponse, HttpResponseBadRequest
from django.urls import reverse_lazy
from django.utils.html import escape

from core_composer_app.components.bucket import api as bucket_api
from core_composer_app.components.bucket.models import Bucket
from core_composer_app.components.type.models import Type
from core_composer_app.components.type_version_manager import (
    api as type_version_manager_api,
)
from core_composer_app.components.type_version_manager.models import TypeVersionManager
from core_composer_app.views.admin.forms import EditBucketForm
from core_main_app.commons import exceptions
from core_main_app.components.template.api import init_template_with_dependencies
from core_main_app.components.version_manager import api as version_manager_api
from core_main_app.views.admin.ajax import (
    _get_xsd_content_from_html,
    _get_dependencies_dict,
)
from core_main_app.views.common.ajax import EditObjectModalView


@staff_member_required
def delete_bucket(request):
    """Delete a bucket.

    Args:
        request:

    Returns:

    """
    try:
        bucket_id = request.POST["bucket_id"]

        bucket = bucket_api.get_by_id(bucket_id)
        bucket_api.delete(bucket)
    except Exception as e:
        return HttpResponseBadRequest(escape(str(e)))

    return HttpResponse(json.dumps({}), content_type="application/javascript")


@staff_member_required
def resolve_dependencies(request):
    """Resolve import/includes to avoid local references.

    Args:
        request:

    Returns:

    """
    try:
        # Get the parameters
        name = request.POST.get("name", None)
        version_manager_id = request.POST.get("version_manager_id", "")
        filename = request.POST["filename"]
        xsd_content = request.POST["xsd_content"]
        schema_locations = request.POST.getlist("schemaLocations[]")
        dependencies = request.POST.getlist("dependencies[]")
        buckets = request.POST.getlist("buckets[]")

        # create new object
        type_object = Type(
            filename=filename, content=_get_xsd_content_from_html(xsd_content)
        )
        init_template_with_dependencies(
            type_object,
            _get_dependencies_dict(schema_locations, dependencies),
            request=request,
        )

        # get the version manager or create a new one
        if version_manager_id != "":
            type_version_manager = version_manager_api.get(
                version_manager_id, request=request
            )
        else:
            type_version_manager = TypeVersionManager(title=name)
        type_version_manager_api.insert(
            type_version_manager, type_object, request=request, list_bucket_ids=buckets
        )
    except Exception as e:
        return HttpResponseBadRequest(
            escape(str(e)), content_type="application/javascript"
        )

    return HttpResponse(json.dumps({}), content_type="application/javascript")


class EditBucketView(EditObjectModalView):
    form_class = EditBucketForm
    document = Bucket
    success_url = reverse_lazy("admin:core_composer_app_buckets")
    success_message = "Label edited with success."

    def _save(self, form):
        # Save treatment.
        try:
            self.object.label = form.cleaned_data.get("label")
            bucket_api.upsert(self.object)
        except exceptions.NotUniqueError:
            form.add_error(
                None,
                "A bucket with the same label already exists. Please choose another label.",
            )
        except Exception as e:
            form.add_error(None, str(e))



"""Composer app user views
"""
from os.path import join

from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.contrib.staticfiles import finders
from django.urls import reverse_lazy

from core_composer_app.components.bucket import api as bucket_api
from core_composer_app.components.type_version_manager import (
    api as type_version_manager_api,
)
from core_composer_app.permissions import rights
from core_main_app.components.template import api as template_api
from core_main_app.components.template_version_manager import (
    api as template_version_manager_api,
)
from core_main_app.components.version_manager import api as version_manager_api
from core_main_app.utils import decorators as decorators
from core_main_app.utils.file import read_file_content, get_file_http_response
from core_main_app.utils.rendering import render
from core_main_app.utils.xml import xsl_transform
from core_main_app.views.user.views import get_context_manage_template_versions
from xml_utils.commons.constants import LXML_SCHEMA_NAMESPACE
from xml_utils.xsd_tree.operations.annotation import remove_annotations
from xml_utils.xsd_tree.xsd_tree import XSDTree
from xml_utils.xsd_types.xsd_types import get_xsd_types


# TODO: see if sessions are problematic


@decorators.permission_required(
    content_type=rights.composer_content_type,
    permission=rights.composer_access,
    login_url=reverse_lazy("core_main_app_login"),
)
def index(request):
    """Page that allows to select a template to start composing.

    Args:
        request:

    Returns:

    """
    assets = {"js": [], "css": []}

    global_active_template_list = (
        template_version_manager_api.get_active_global_version_manager(
            request=request, _cls=True
        )
    )
    user_active_template_list = (
        template_version_manager_api.get_active_version_manager_by_user_id(
            request=request, _cls=True
        )
    )

    global_active_type_list = (
        type_version_manager_api.get_active_global_version_manager(request=request)
    )
    user_active_type_list = (
        type_version_manager_api.get_active_version_manager_by_user_id(request=request)
    )

    context = {
        "global_templates": global_active_template_list,
        "global_types": global_active_type_list,
        "user_templates": user_active_template_list,
        "user_types": user_active_type_list,
    }

    return render(
        request, "core_composer_app/user/index.html", assets=assets, context=context
    )


@decorators.permission_required(
    content_type=rights.composer_content_type,
    permission=rights.composer_access,
    login_url=reverse_lazy("core_main_app_login"),
)
def build_template(request, template_id):
    """View that allows to build the Template.

    Args:
        request:
        template_id:

    Returns:

    """
    if template_id == "new":
        base_template_path = finders.find(
            join("core_composer_app", "user", "xsd", "new_base_template.xsd")
        )
        xsd_string = read_file_content(base_template_path)
    else:
        template = template_api.get(template_id, request=request)
        xsd_string = template.content

    request.session["newXmlTemplateCompose"] = xsd_string
    request.session["includedTypesCompose"] = []

    # store the current includes/imports
    xsd_tree = XSDTree.build_tree(xsd_string)
    includes = xsd_tree.findall("{}include".format(LXML_SCHEMA_NAMESPACE))
    for el_include in includes:
        if "schemaLocation" in el_include.attrib:
            request.session["includedTypesCompose"].append(
                el_include.attrib["schemaLocation"]
            )
    imports = xsd_tree.findall("{}import".format(LXML_SCHEMA_NAMESPACE))
    for el_import in imports:
        if "schemaLocation" in el_import.attrib:
            request.session["includedTypesCompose"].append(
                el_import.attrib["schemaLocation"]
            )

    # remove annotations from the tree
    remove_annotations(xsd_tree)
    xsd_string = XSDTree.tostring(xsd_tree)

    # loads XSLT
    xslt_path = finders.find(join("core_composer_app", "user", "xsl", "xsd2html.xsl"))
    # reads XSLT
    xslt_string = read_file_content(xslt_path)
    # transform XML to HTML
    xsd_to_html_string = xsl_transform(xsd_string, xslt_string)

    # 1) Get user defined types
    user_types = type_version_manager_api.get_version_managers_by_user(request=request)
    # 2) Get buckets
    buckets = bucket_api.get_all()

    # 3) no_buckets_types: list of types that are not assigned to a specific bucket
    no_buckets_types = type_version_manager_api.get_no_buckets_types(request=request)

    # 4) Build list of built-in types
    built_in_types = []
    for built_in_type in get_xsd_types():
        built_in_types.append({"current": "built_in_type", "title": built_in_type})

    assets = {
        "js": [
            {"path": "core_composer_app/user/js/build_template.js", "is_raw": False},
            {"path": "core_composer_app/user/js/build_template.raw.js", "is_raw": True},
            {"path": "core_composer_app/user/js/xpath.js", "is_raw": False},
            {"path": "core_composer_app/user/js/menus.js", "is_raw": False},
            {"path": "core_composer_app/user/js/xsd_tree.js", "is_raw": False},
        ],
        "css": [
            "core_main_app/common/css/XMLTree.css",
            "core_composer_app/common/css/bucket.css",
            "core_composer_app/user/css/menu.css",
            "core_composer_app/user/css/style.css",
            "core_composer_app/user/css/xsd_tree.css",
        ],
    }
    context = {
        "buckets": buckets,
        "built_in_types": built_in_types,
        "no_buckets_types": no_buckets_types,
        "user_types": user_types,
        "xsd_form": xsd_to_html_string,
        "template_id": template_id,
    }

    modals = [
        "core_composer_app/user/builder/menus/sequence.html",
        "core_composer_app/user/builder/menus/element.html",
        "core_composer_app/user/builder/menus/element_root.html",
        "core_composer_app/user/builder/modals/root_type_name.html",
        "core_composer_app/user/builder/modals/element_name.html",
        "core_composer_app/user/builder/modals/insert_element.html",
        "core_composer_app/user/builder/modals/delete_element.html",
        "core_composer_app/user/builder/modals/change_type.html",
        "core_composer_app/user/builder/modals/save_template.html",
        "core_composer_app/user/builder/modals/save_type.html",
        "core_composer_app/user/builder/modals/save_success.html",
        "core_composer_app/user/builder/modals/occurrences.html",
        "core_composer_app/user/builder/modals/errors.html",
    ]

    return render(
        request,
        "core_composer_app/user/build_template.html",
        assets=assets,
        context=context,
        modals=modals,
    )


@decorators.permission_required(
    content_type=rights.composer_content_type,
    permission=rights.composer_access,
    login_url=reverse_lazy("core_main_app_login"),
)
def download_xsd(request):
    """Make the current XSD available for download.

    Args:
        request:

    Returns:

    """
    xsd_string = request.session["newXmlTemplateCompose"]

    # return the file
    return get_file_http_response(
        file_content=xsd_string,
        file_name="schema.xsd",
        content_type="application/xsd",
        extension=".xsd",
    )


@login_required
def manage_type_versions(request, version_manager_id):
    """View that allows type versions management.

    Args:
        request:
        version_manager_id:

    Returns:

    """
    try:
        # get the version manager
        version_manager = version_manager_api.get(version_manager_id, request=request)
        context = get_context_manage_template_versions(version_manager, "Type")
        if "core_parser_app" in settings.INSTALLED_APPS:
            context.update({"module_url": "core_composer_app_type_modules"})

        assets = {
            "js": [
                {
                    "path": "core_main_app/common/js/templates/versions/set_current.js",
                    "is_raw": False,
                },
                {
                    "path": "core_main_app/common/js/templates/versions/restore.js",
                    "is_raw": False,
                },
                {
                    "path": "core_main_app/common/js/templates/versions/modals/disable.js",
                    "is_raw": False,
                },
            ]
        }

        modals = ["core_main_app/admin/templates/versions/modals/disable.html"]

        return render(
            request,
            "core_composer_app/user/types/versions.html",
            assets=assets,
            modals=modals,
            context=context,
        )
    except Exception as e:
        return render(
            request,
            "core_main_app/common/commons/error.html",
            context={"error": str(e)},
        )



""" Integration Test for Bucket Rest API
"""

from rest_framework import status

from core_composer_app.rest.bucket import views
from core_main_app.utils.integration_tests.integration_base_test_case import (
    MongoIntegrationBaseTestCase,
)
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from core_main_app.utils.tests_tools.RequestMock import RequestMock
from tests.components.bucket.fixtures.fixtures import BucketFixtures

fixture_bucket = BucketFixtures()


class TestBucketList(MongoIntegrationBaseTestCase):
    fixture = fixture_bucket

    def setUp(self):
        super(TestBucketList, self).setUp()
        self.data = {"label": "label"}

    def test_get_returns_http_200(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(views.BucketList.as_view(), user)

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_get_returns_all_buckets(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(views.BucketList.as_view(), user)

        # Assert
        self.assertEqual(len(response.data), 3)

    def test_get_by_correct_label_returns_one_bucket(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            views.BucketList.as_view(),
            user,
            data={"label": self.fixture.bucket_1.label},
        )

        # Assert
        self.assertEqual(len(response.data), 1)

    def test_get_by_correct_label_returns_correct_bucket(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            views.BucketList.as_view(),
            user,
            data={"label": self.fixture.bucket_1.label},
        )

        # Assert
        self.assertEqual(response.data[0]["label"], self.fixture.bucket_1.label)

    def test_get_by_incorrect_label_returns_no_bucket(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            views.BucketList.as_view(), user, data={"label": "incorrect label"}
        )

        # Assert
        self.assertEqual(len(response.data), 0)

    def test_post_returns_http_201(self):
        # Arrange
        user = create_mock_user("1", is_staff=True)

        # Act
        response = RequestMock.do_request_post(
            views.BucketList.as_view(), user, data=self.data
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_post_as_user_returns_http_403(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_post(
            views.BucketList.as_view(), user, data=self.data
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_post_empty_types_returns_http_201(self):
        # Arrange
        user = create_mock_user("1", is_staff=True)
        self.data["types"] = []

        # Act
        response = RequestMock.do_request_post(
            views.BucketList.as_view(), user, data=self.data
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_post_with_one_correct_type_returns_http_201(self):
        # Arrange
        user = create_mock_user("1", is_staff=True)
        self.data["types"] = [str(self.fixture.type_vm_1.id)]

        # Act
        response = RequestMock.do_request_post(
            views.BucketList.as_view(), user, data=self.data
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_post_with_one_correct_type_returns_response_containing_one_type(self):
        # Arrange
        user = create_mock_user("1", is_staff=True)
        self.data["types"] = [str(self.fixture.type_vm_1.id)]

        # Act
        response = RequestMock.do_request_post(
            views.BucketList.as_view(), user, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data["types"]), 1)

    def test_post_with_two_correct_type_returns_http_201(self):
        # Arrange
        user = create_mock_user("1", is_staff=True)
        self.data["types"] = [
            str(self.fixture.type_vm_1.id),
            str(self.fixture.type_vm_2.id),
        ]

        # Act
        response = RequestMock.do_request_post(
            views.BucketList.as_view(), user, data=self.data
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_post_with_one_correct_type_returns_response_containing_two_types(self):
        # Arrange
        user = create_mock_user("1", is_staff=True)
        self.data["types"] = [
            str(self.fixture.type_vm_1.id),
            str(self.fixture.type_vm_2.id),
        ]

        # Act
        response = RequestMock.do_request_post(
            views.BucketList.as_view(), user, data=self.data
        )

        # Assert
        self.assertEqual(len(response.data["types"]), 2)

    def test_post_with_one_incorrect_type_returns_http_400(self):
        # Arrange
        user = create_mock_user("1", is_staff=True)
        self.data["types"] = ["wrong_type_id"]

        # Act
        response = RequestMock.do_request_post(
            views.BucketList.as_view(), user, data=self.data
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)


class TestBucketDetail(MongoIntegrationBaseTestCase):
    fixture = fixture_bucket

    def setUp(self):
        super(TestBucketDetail, self).setUp()
        self.data = {"label": "label"}

    def test_get_returns_http_200(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            views.BucketDetail.as_view(),
            user,
            param={"pk": str(self.fixture.bucket_1.id)},
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_get_returns_bucket(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            views.BucketDetail.as_view(),
            user,
            param={"pk": str(self.fixture.bucket_1.id)},
        )

        # Assert
        self.assertEqual(response.data["label"], self.fixture.bucket_1.label)

    def test_get_wrong_id_returns_http_404(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            views.BucketDetail.as_view(), user, param={"pk": "507f1f77bcf86cd799439011"}
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_delete_returns_http_204(self):
        # Arrange
        user = create_mock_user("1", is_staff=True)

        # Act
        response = RequestMock.do_request_delete(
            views.BucketDetail.as_view(),
            user,
            param={"pk": str(self.fixture.bucket_1.id)},
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_delete_wrong_id_returns_http_404(self):
        # Arrange
        user = create_mock_user("1", is_staff=True)

        # Act
        response = RequestMock.do_request_delete(
            views.BucketDetail.as_view(), user, param={"pk": "507f1f77bcf86cd799439011"}
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_patch_returns_http_200(self):
        # Arrange
        user = create_mock_user("1", is_staff=True)

        # Act
        response = RequestMock.do_request_patch(
            views.BucketDetail.as_view(),
            user,
            param={"pk": str(self.fixture.bucket_1.id)},
            data=self.data,
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_patch_returns_updated_label(self):
        # Arrange
        user = create_mock_user("1", is_staff=True)

        # Act
        response = RequestMock.do_request_patch(
            views.BucketDetail.as_view(),
            user,
            param={"pk": str(self.fixture.bucket_1.id)},
            data=self.data,
        )

        # Assert
        self.assertEqual(response.data["label"], self.data["label"])

    def test_patch_wrong_id_returns_http_404(self):
        # Arrange
        user = create_mock_user("1", is_staff=True)

        # Act
        response = RequestMock.do_request_patch(
            views.BucketDetail.as_view(), user, param={"pk": "507f1f77bcf86cd799439011"}
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)


class TestTypeVersionManagerBuckets(MongoIntegrationBaseTestCase):
    fixture = fixture_bucket

    def setUp(self):
        super(TestTypeVersionManagerBuckets, self).setUp()

    def test_patch_returns_http_200(self):
        # Arrange
        user = create_mock_user("1", is_staff=True)
        self.data = [{"id": str(self.fixture.bucket_1.id)}]

        # Act
        response = RequestMock.do_request_patch(
            views.TypeVersionManagerBuckets.as_view(),
            user,
            param={"pk": self.fixture.type_vm_1.id},
            data=self.data,
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_patch_two_correct_bucket_id_returns_http_200(self):
        # Arrange
        user = create_mock_user("1", is_staff=True)
        self.data = [
            {"id": str(self.fixture.bucket_1.id)},
            {"id": str(self.fixture.bucket_2.id)},
        ]

        # Act
        response = RequestMock.do_request_patch(
            views.TypeVersionManagerBuckets.as_view(),
            user,
            param={"pk": self.fixture.type_vm_1.id},
            data=self.data,
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_patch_no_bucket_id_returns_http_200(self):
        # Arrange
        user = create_mock_user("1", is_staff=True)
        self.data = []

        # Act
        response = RequestMock.do_request_patch(
            views.TypeVersionManagerBuckets.as_view(),
            user,
            param={"pk": self.fixture.type_vm_1.id},
            data=self.data,
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_patch_wrong_bucket_id_returns_http_404(self):
        # Arrange
        user = create_mock_user("1", is_staff=True)
        self.data = [{"id": "507f1f77bcf86cd799439011"}]

        # Act
        response = RequestMock.do_request_patch(
            views.TypeVersionManagerBuckets.as_view(),
            user,
            param={"pk": self.fixture.type_vm_1.id},
            data=self.data,
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_patch_wrong_template_version_manager_id_returns_http_404(self):
        # Arrange
        user = create_mock_user("1", is_staff=True)
        self.data = [{"id": str(self.fixture.bucket_1.id)}]

        # Act
        response = RequestMock.do_request_patch(
            views.TypeVersionManagerBuckets.as_view(),
            user,
            param={"pk": "507f1f77bcf86cd799439011"},
            data=self.data,
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_patch_other_user_type_version_manager_returns_http_403(self):
        # Arrange
        user = create_mock_user("2")
        self.data = [{"id": str(self.fixture.bucket_1.id)}]

        # Act
        response = RequestMock.do_request_patch(
            views.TypeVersionManagerBuckets.as_view(),
            user,
            param={"pk": self.fixture.type_vm_1.id},
            data=self.data,
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)



"""XML utils for Composer app
"""
from core_main_app.commons.exceptions import CoreError, XMLError
from core_main_app.utils.xml import is_well_formed_xml, validate_xml_schema

from xml_utils.commons.constants import LXML_SCHEMA_NAMESPACE
from xml_utils.xsd_tree.operations.namespaces import (
    get_namespaces,
    get_default_prefix,
    get_target_namespace,
)
from xml_utils.xsd_tree.xsd_tree import XSDTree

COMPLEX_TYPE = "complexType"
SIMPLE_TYPE = "simpleType"


def check_type_core_support(xsd_string):
    """Check that the format of the the type is supported by the current version of the Core.
    Return the type definition (simpleType or complexType).

    Args:
        xsd_string:

    Returns:
        type_definition: simpleType or complexType.

    """
    type_definition = ""
    error_message = (
        "A type should be a valid XML schema containing only one type definition "
        "(Allowed tags are: simpleType or complexType and include)."
    )

    # check that well formatted first
    if not is_well_formed_xml(xsd_string):
        raise XMLError("Uploaded file is not well formatted XML.")

    # build the tree
    xsd_tree = XSDTree.build_tree(xsd_string)

    # get elements
    elements = xsd_tree.findall("*")

    if len(elements) > 0:
        # only simpleType, complexType or include
        for element in elements:
            if (
                "complexType" not in element.tag
                and "simpleType" not in element.tag
                and "include" not in element.tag
            ):
                raise CoreError(error_message)

        # only one type
        cpt_type = 0
        for element in elements:
            if "complexType" in element.tag or "simpleType" in element.tag:
                cpt_type += 1
                if cpt_type > 1:
                    raise CoreError(error_message)
                type_definition = (
                    COMPLEX_TYPE if "complexType" in element.tag else SIMPLE_TYPE
                )
    else:
        raise CoreError(error_message)

    return type_definition


def remove_single_root_element(xsd_string):
    """Remove root element from the xsd string.

    Args:
        xsd_string:

    Returns:

    """
    # Build xsd tree
    xsd_tree = XSDTree.build_tree(xsd_string)
    # find the root element
    root = xsd_tree.find("{}element".format(LXML_SCHEMA_NAMESPACE))
    if root is not None:
        # remove root element from parent (schema)
        root.getparent().remove(root)
        # convert the tree to back string
        xsd_string = XSDTree.tostring(xsd_tree)
    # return xsd string
    return xsd_string


def rename_single_root_type(xsd_string, type_name):
    """Rename the type of the single root element.

    Args:
        xsd_string:
        type_name:

    Returns:

    """
    # build xsd tree
    xsd_tree = XSDTree.build_tree(xsd_string)
    # xpath to the single root element
    xpath_root = LXML_SCHEMA_NAMESPACE + "element"
    # xpath to the single root type
    xpath_root_type = LXML_SCHEMA_NAMESPACE + "complexType"
    # change the root type name in the xsd tree
    xsd_tree.find(xpath_root).attrib["type"] = type_name
    xsd_tree.find(xpath_root_type).attrib["name"] = type_name
    # rebuild xsd string
    xsd_string = XSDTree.tostring(xsd_tree)
    # return xsd string
    return xsd_string


def delete_xsd_element(xsd_string, xpath):
    """Delete element from tree.

    Args:
        xsd_string:
        xpath:

    Returns:

    """
    # build xsd tree
    xsd_tree = XSDTree.build_tree(xsd_string)
    # get xsd namespaces
    namespaces = get_namespaces(xsd_string)
    # get default prefix
    default_prefix = get_default_prefix(namespaces)
    # set the element namespace
    xpath = xpath.replace(default_prefix + ":", LXML_SCHEMA_NAMESPACE)
    # get element to remove from tree
    element_to_remove = xsd_tree.find(xpath)
    # remove element from tree
    element_to_remove.getparent().remove(element_to_remove)

    # rebuild xsd string
    xsd_string = XSDTree.tostring(xsd_tree)
    # return xsd string
    return xsd_string


def change_xsd_element_type(xsd_string, xpath, type_name):
    """Change the type of an element (e.g. sequence -> choice).

    Args:
        xsd_string:
        xpath:
        type_name:

    Returns:

    """
    xsd_tree = XSDTree.build_tree(xsd_string)
    namespaces = get_namespaces(xsd_string)
    default_prefix = get_default_prefix(namespaces)

    # set the element namespace
    xpath = xpath.replace(default_prefix + ":", LXML_SCHEMA_NAMESPACE)
    xsd_tree.find(xpath).tag = LXML_SCHEMA_NAMESPACE + type_name

    # rebuild xsd string
    xsd_string = XSDTree.tostring(xsd_tree)

    # return xsd string
    return xsd_string


def set_xsd_element_occurrences(xsd_string, xpath, min_occurs, max_occurs):
    """Set occurrences of element.

    Args:
        xsd_string:
        xpath:
        min_occurs:
        max_occurs:

    Returns:

    """
    # build xsd tree
    xsd_tree = XSDTree.build_tree(xsd_string)
    # get namespaces
    namespaces = get_namespaces(xsd_string)
    # get default prefix
    default_prefix = get_default_prefix(namespaces)

    # set the element namespace
    xpath = xpath.replace(default_prefix + ":", LXML_SCHEMA_NAMESPACE)
    # add the element to the sequence
    element = xsd_tree.find(xpath)
    element.attrib["minOccurs"] = min_occurs
    element.attrib["maxOccurs"] = max_occurs

    # save the tree in the session
    xsd_string = XSDTree.tostring(xsd_tree)

    # return xsd string
    return xsd_string


def get_xsd_element_occurrences(xsd_string, xpath):
    """Get the min and max occurrences of the element.

    Args:
        xsd_string:
        xpath:

    Returns:

    """
    # build the xsd tree
    xsd_tree = XSDTree.build_tree(xsd_string)
    # get the namespaces
    namespaces = get_namespaces(xsd_string)
    # get the default prefix
    default_prefix = get_default_prefix(namespaces)

    # set the element namespace
    xpath = xpath.replace(default_prefix + ":", LXML_SCHEMA_NAMESPACE)
    # add the element to the sequence
    element = xsd_tree.find(xpath)

    if "minOccurs" in element.attrib:
        min_occurs = element.attrib["minOccurs"]
    else:
        min_occurs = "1"
    if "maxOccurs" in element.attrib:
        max_occurs = element.attrib["maxOccurs"]
    else:
        max_occurs = "1"

    return min_occurs, max_occurs


def rename_xsd_element(xsd_string, xpath, new_name):
    """Rename xsd element.

    Args:
        xsd_string:
        xpath:
        new_name:

    Returns:

    """
    # build the xsd tree
    xsd_tree = XSDTree.build_tree(xsd_string)
    # get the namespaces
    namespaces = get_namespaces(xsd_string)
    # get the default prefix
    default_prefix = get_default_prefix(namespaces)

    # set the element namespace
    xpath = xpath.replace(default_prefix + ":", LXML_SCHEMA_NAMESPACE)
    # rename element
    xsd_tree.find(xpath).attrib["name"] = new_name

    # rebuild xsd string
    xsd_string = XSDTree.tostring(xsd_tree)
    # return xsd string
    return xsd_string


# TODO: refactor more
def _insert_element_type(
    xsd_string, xpath, type_content, element_type_name, include_url
):
    """Insert an element of given type in xsd string.

    Args:
        xsd_string: xsd string
        xpath: xpath where to insert the element
        type_content: string content of the type to insert
        element_type_name: name of the type
        include_url: url used to reference the type in schemaLocation

    Returns:

    """
    # build the dom tree of the schema being built
    xsd_tree = XSDTree.build_tree(xsd_string)
    # get namespaces information for the schema
    namespaces = get_namespaces(xsd_string)
    # get the default namespace
    default_prefix = get_default_prefix(namespaces)
    # get target namespace information
    target_namespace, target_namespace_prefix = get_target_namespace(
        xsd_tree, namespaces
    )
    # build xpath to element
    xpath = xpath.replace(default_prefix + ":", LXML_SCHEMA_NAMESPACE)
    # build xsd tree
    type_xsd_tree = XSDTree.build_tree(type_content)
    # get namespaces information for the type
    type_namespaces = get_namespaces(type_content)
    # get target namespace information
    type_target_namespace, type_target_namespace_prefix = get_target_namespace(
        type_xsd_tree, type_namespaces
    )

    # get the type from the included/imported file
    # If there is a complex type
    element_type = type_xsd_tree.find("{}complexType".format(LXML_SCHEMA_NAMESPACE))
    if element_type is None:
        # If there is a simple type
        element_type = type_xsd_tree.find("{}simpleType".format(LXML_SCHEMA_NAMESPACE))
    type_name = element_type.attrib["name"]

    # format type name to avoid forbidden xml characters
    element_type_name = _get_valid_xml_name(element_type_name)

    # variable that indicates if namespaces map needs to be updated
    update_ns_map = False

    # Schema without target namespace
    if target_namespace is None:
        # Type without target namespace
        if type_target_namespace is None:
            # create type name with namespace
            ns_type_name = type_name
            # create include element
            dependency_tag = "include"
            dependency_attrib = {"schemaLocation": include_url}
        # Type with target namespace
        else:
            # create type name with namespace
            ns_type_name = _get_ns_type_name(
                type_target_namespace_prefix, type_name, prefix_required=True
            )
            # create import element
            dependency_tag = "import"
            dependency_attrib = {
                "schemaLocation": include_url,
                "namespace": type_target_namespace,
            }
            update_ns_map = True

    # Schema with target namespace
    else:
        # Type without target namespace
        if type_target_namespace is None:
            # create type name with namespace
            ns_type_name = _get_ns_type_name(target_namespace_prefix, type_name)
            # create include element
            dependency_tag = "include"
            dependency_attrib = {"schemaLocation": include_url}
        # Type with target namespace
        else:
            # Same target namespace as base template
            if target_namespace == type_target_namespace:
                # create type name with namespace
                ns_type_name = _get_ns_type_name(target_namespace_prefix, type_name)
                # create include element
                dependency_tag = "include"
                dependency_attrib = {"schemaLocation": include_url}
            # Different target namespace as base template
            else:
                # create type name with namespace
                ns_type_name = _get_ns_type_name(
                    type_target_namespace_prefix, type_name, prefix_required=True
                )
                # create import element
                dependency_tag = "import"
                dependency_attrib = {
                    "schemaLocation": include_url,
                    "namespace": type_target_namespace,
                }
                update_ns_map = True

    # create dependency element
    dependency_element = _create_xsd_element(dependency_tag, dependency_attrib)
    # create xsd element
    xsd_element = _create_xsd_element(
        "element", attrib={"name": element_type_name, "type": ns_type_name}
    )
    # check if dependency element (include/import) is already present
    dependency_tag = "{0}[@schemaLocation='{1}']".format(
        dependency_element.tag, dependency_element.attrib["schemaLocation"]
    )
    dependency_present = xsd_tree.find(dependency_tag) is not None

    if not dependency_present:
        # add dependency element (include/import)
        xsd_tree.getroot().insert(0, dependency_element)

    # add xsd element
    xsd_tree.find(xpath).append(xsd_element)

    # if namespace map of the schema needs to be updated
    if not dependency_present and update_ns_map:
        root = xsd_tree.getroot()
        root_ns_map = root.nsmap

        if (
            type_target_namespace_prefix in list(root_ns_map.keys())
            and root_ns_map[type_target_namespace_prefix] != type_target_namespace
        ):
            raise CoreError(
                "The namespace prefix is already declared for a different namespace."
            )
        else:
            root_ns_map[type_target_namespace_prefix] = type_target_namespace
            new_root = XSDTree.create_element(
                root.tag, nsmap=root_ns_map, attrib=root.attrib
            )
            new_root[:] = root[:]

            # return result tree
            return new_root

    else:
        # return result tree
        return xsd_tree


def insert_element_type(
    xsd_string, xpath, type_content, element_type_name, include_url, request
):
    """Insert an element of given type in xsd string, and validates result.

    Args:
        xsd_string: xsd string
        xpath: xpath where to insert the element
        type_content: string content of the type to insert
        element_type_name: name of the type
        include_url: url used to reference the type in schemaLocation
        request: request

    Returns:

    """

    new_xsd_tree = _insert_element_type(
        xsd_string, xpath, type_content, element_type_name, include_url
    )
    error = validate_xml_schema(new_xsd_tree, request=request)

    # if errors, raise exception
    if error is not None:
        raise XMLError(error)

    new_xsd_string = XSDTree.tostring(new_xsd_tree)

    return new_xsd_string


def insert_element_built_in_type(xsd_string, xpath, element_type_name, request):
    """Insert element with a builtin type in xsd string.

    Args:
        xsd_string: xsd string
        xpath: xpath where to insert the element
        element_type_name: name of the type to insert
        request: request

    Returns:

    """
    # build the dom tree of the schema being built
    xsd_tree = XSDTree.build_tree(xsd_string)
    # get namespaces information for the schema
    namespaces = get_namespaces(xsd_string)
    # get the default namespace
    default_prefix = get_default_prefix(namespaces)
    # build xpath to element
    xpath = xpath.replace(default_prefix + ":", LXML_SCHEMA_NAMESPACE)

    type_name = default_prefix + ":" + element_type_name
    xsd_tree.find(xpath).append(
        XSDTree.create_element(
            "{}element".format(LXML_SCHEMA_NAMESPACE),
            attrib={"type": type_name, "name": element_type_name},
        )
    )
    # validate XML schema
    error = validate_xml_schema(xsd_tree, request=request)

    # if errors, raise exception
    if error is not None:
        raise XMLError(error)

    return XSDTree.tostring(xsd_tree)


def _get_ns_type_name(prefix, type_name, prefix_required=False):
    """Return type name formatted with namespace prefix.

    Args:
        prefix:
        type_name:

    Returns:

    """
    if prefix != "":
        ns_type_name = "{0}:{1}".format(prefix, type_name)
    else:
        if not prefix_required:
            ns_type_name = type_name
        else:
            raise CoreError(
                "Unable to add the type because no prefix is defined for the target namespace."
            )

    return ns_type_name


def _get_valid_xml_name(name):
    """Return a valid xml name.

    Args:
        name:

    Returns:

    """

    default_name = "default_name"
    try:
        # not allowed to start with XML
        if name.upper().startswith("XML"):
            name = "_{}".format(name)

        # not allowed to start with something other than letter or underscore
        if not name[0].isalpha() or name[0] == "_":
            name = "_{}".format(name)

        # not allowed to have spaces
        name = name.replace(" ", "")

        if len(name) == 0:
            name = default_name
    except:
        name = default_name

    return name


def _create_xsd_element(tag, attrib):
    """Create an XSD element.

    Args:
        tag:
        attrib:

    Returns:

    """

    if tag not in ["element", "include", "import"]:
        raise CoreError("Unable to create XSD element: invalid tag")

    xsd_element = XSDTree.create_element(
        "{0}{1}".format(LXML_SCHEMA_NAMESPACE, tag), attrib=attrib
    )

    return xsd_element



""" Apps file for setting core package when app is ready
"""
import sys

from django.apps import AppConfig


class ComposerAppConfig(AppConfig):
    """Core composer application settings"""

    name = "core_composer_app"

    def ready(self):
        """Run when the app is ready.

        Returns:

        """
        if "migrate" not in sys.argv:
            import core_composer_app.permissions.discover as discover

            discover.init_permissions()



"""Type Version Manager test cases
"""
from unittest.case import TestCase

from bson.objectid import ObjectId
from django.core import exceptions as django_exceptions
from django.test import override_settings
from mock.mock import Mock, patch
from mongoengine import errors as mongoengine_exceptions

from core_composer_app.components.bucket.models import Bucket
from core_composer_app.components.type.models import Type
from core_composer_app.components.type_version_manager import api as version_manager_api
from core_composer_app.components.type_version_manager.api import get_no_buckets_types
from core_composer_app.components.type_version_manager.models import TypeVersionManager
from core_main_app.commons.exceptions import CoreError, ModelError, NotUniqueError
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from core_main_app.utils.tests_tools.RequestMock import create_mock_request


class TestTypeVersionManagerInsert(TestCase):
    @patch.object(TypeVersionManager, "save")
    @patch.object(Type, "save")
    def test_create_version_manager_raises_exception_if_type_not_supported(
        self, mock_save_type, mock_save_type_version_manager
    ):
        # Arrange
        mock_user = create_mock_user("1", is_superuser=True)
        mock_request = create_mock_request(user=mock_user)
        type_filename = "schema.xsd"
        type_content = "<schema xmlns='http://www.w3.org/2001/XMLSchema'></schema>"
        type_object = _create_type(type_filename, type_content)

        mock_save_type.return_value = type_object

        version_manager = _create_type_version_manager(title="Schema")
        mock_save_type_version_manager.return_value = version_manager

        # Act + Assert
        with self.assertRaises(CoreError):
            version_manager_api.insert(
                version_manager, type_object, request=mock_request
            )

    @override_settings(ROOT_URLCONF="core_main_app.urls")
    @patch.object(TypeVersionManager, "save")
    @patch.object(Type, "save")
    def test_create_version_manager_returns_version_manager(
        self, mock_save_type, mock_save_type_version_manager
    ):
        # Arrange
        mock_user = create_mock_user("1", is_superuser=True)
        mock_request = create_mock_request(user=mock_user)
        type_filename = "schema.xsd"
        type_content = (
            "<schema xmlns='http://www.w3.org/2001/XMLSchema'><simpleType name='type'>"
            "<restriction base='string'><enumeration value='test'/></restriction>"
            "</simpleType></schema>"
        )
        type_object = _create_type(type_filename, type_content)

        mock_save_type.return_value = type_object

        version_manager = _create_type_version_manager(title="Schema")
        mock_save_type_version_manager.return_value = version_manager

        # Act
        result = version_manager_api.insert(
            version_manager, type_object, request=mock_request
        )

        # Assert
        self.assertIsInstance(result, TypeVersionManager)

    @override_settings(ROOT_URLCONF="core_main_app.urls")
    @patch.object(Type, "delete")
    @patch.object(Type, "save")
    @patch.object(TypeVersionManager, "save")
    def test_insert_manager_raises_api_error_if_title_already_exists(
        self, mock_version_manager_save, mock_save, mock_delete
    ):
        # Arrange
        mock_user = create_mock_user("1", is_superuser=True)
        mock_request = create_mock_request(user=mock_user)
        type_filename = "schema.xsd"
        type_object = _create_type(type_filename)

        mock_save.return_value = type_object
        mock_delete.return_value = None
        mock_version_manager = _create_type_version_manager(title="Schema")
        mock_version_manager_save.side_effect = mongoengine_exceptions.NotUniqueError

        # Act + Assert
        with self.assertRaises(NotUniqueError):
            version_manager_api.insert(
                mock_version_manager, type_object, request=mock_request
            )

    @override_settings(ROOT_URLCONF="core_main_app.urls")
    @patch.object(Type, "save")
    def test_create_version_manager_raises_exception_if_error_in_create_type(
        self, mock_save
    ):
        # Arrange
        mock_user = create_mock_user("1", is_superuser=True)
        mock_request = create_mock_request(user=mock_user)
        type_filename = "schema.xsd"
        type_object = _create_type(type_filename)

        mock_version_manager = _create_mock_type_version_manager(title="Schema")
        mock_save.side_effect = django_exceptions.ValidationError("")

        # Act + Assert
        with self.assertRaises(django_exceptions.ValidationError):
            version_manager_api.insert(
                mock_version_manager, type_object, request=mock_request
            )

    @override_settings(ROOT_URLCONF="core_main_app.urls")
    @patch.object(Type, "delete")
    @patch.object(TypeVersionManager, "save")
    @patch.object(Type, "save")
    def test_create_version_manager_raises_exception_if_error_in_create_version_manager(
        self, mock_save, mock_save_version_manager, mock_delete
    ):
        # Arrange
        mock_user = create_mock_user("1", is_superuser=True)
        mock_request = create_mock_request(user=mock_user)
        type_filename = "Schema"
        type_object = _create_type(type_filename)

        mock_save.return_value = type_object
        version_manager = _create_type_version_manager(title="Schema")
        mock_save_version_manager.side_effect = django_exceptions.ValidationError("")
        mock_delete.return_value = None

        # Act + Assert
        with self.assertRaises(ModelError):
            version_manager_api.insert(
                version_manager, type_object, request=mock_request
            )


class TestTypeVersionManagerGetGlobalVersions(TestCase):
    @patch.object(TypeVersionManager, "get_global_version_managers")
    def test_get_global_version_managers_returns_types(
        self, mock_get_global_version_managers
    ):
        # Arrange
        mock_user = create_mock_user("1", is_superuser=True)
        mock_request = create_mock_request(user=mock_user)
        mock_type1 = _create_mock_type()
        mock_type2 = _create_mock_type()

        mock_get_global_version_managers.return_value = [mock_type1, mock_type2]

        result = version_manager_api.get_global_version_managers(request=mock_request)

        # Assert
        self.assertTrue(all(isinstance(item, Type) for item in result))


class TestTypeVersionManagerGetVersionManagersByUser(TestCase):
    @patch.object(TypeVersionManager, "get_version_managers_by_user")
    def test_get_version_managers_by_user_returns_types_with_given_user_id(
        self, mock_get_version_managers_by_user
    ):
        # Arrange
        user_id = "10"
        mock_user = create_mock_user(user_id=user_id, is_superuser=True)
        mock_request = create_mock_request(user=mock_user)
        mock_type1 = _create_mock_type_version_manager(user=user_id)
        mock_type2 = _create_mock_type_version_manager(user=user_id)

        mock_get_version_managers_by_user.return_value = [mock_type1, mock_type2]

        result = version_manager_api.get_version_managers_by_user(request=mock_request)

        # Assert
        self.assertTrue(item.user_id == user_id for item in result)


class TestGetNoBucketsTypes(TestCase):
    @patch.object(TypeVersionManager, "get_global_version_managers")
    @patch.object(Bucket, "get_all")
    def test_get_no_buckets_types_returns_types(
        self, mock_get_all_buckets, mock_get_global_version_managers
    ):
        # Arrange
        mock_user = create_mock_user("1", is_superuser=True)
        mock_request = create_mock_request(user=mock_user)
        mock_type1 = _create_mock_type_version_manager()
        mock_type2 = _create_mock_type_version_manager()
        mock_get_global_version_managers.return_value = [mock_type1, mock_type2]
        mock_get_all_buckets.return_value = []

        result = get_no_buckets_types(request=mock_request)
        self.assertTrue(all(isinstance(item, TypeVersionManager) for item in result))

    @patch.object(TypeVersionManager, "get_global_version_managers")
    @patch.object(Bucket, "get_all")
    def test_get_no_buckets_types_returns_all_types_if_no_buckets(
        self, mock_get_all_buckets, mock_get_global_version_managers
    ):
        # Arrange
        mock_user = create_mock_user("1", is_superuser=True)
        mock_request = create_mock_request(user=mock_user)
        mock_type1 = _create_mock_type_version_manager()
        mock_type2 = _create_mock_type_version_manager()
        mock_get_global_version_managers.return_value = [mock_type1, mock_type2]
        mock_get_all_buckets.return_value = []

        self.assertTrue(len(get_no_buckets_types(request=mock_request)) == 2)

    @patch.object(TypeVersionManager, "get_global_version_managers")
    @patch.object(Bucket, "get_all")
    def test_get_no_buckets_types_returns_all_types_not_in_buckets(
        self, mock_get_all_buckets, mock_get_global_version_managers
    ):
        # Arrange
        mock_user = create_mock_user("1", is_superuser=True)
        mock_request = create_mock_request(user=mock_user)
        mock_type1 = _create_mock_type_version_manager()
        mock_type2 = _create_mock_type_version_manager()

        mock_bucket = _create_mock_bucket(types=[mock_type1])

        mock_get_global_version_managers.return_value = [mock_type1, mock_type2]
        mock_get_all_buckets.return_value = [mock_bucket]

        self.assertTrue(len(get_no_buckets_types(request=mock_request)) == 1)


def _create_mock_bucket(types=None):
    """Returns a mock bucket

    Args:
        types:

    Returns:

    """
    if types is None:
        types = []
    mock_bucket = Mock(spec=Bucket)
    mock_bucket.label = "bucket"
    mock_bucket.label = "#000000"
    mock_bucket.types = types
    return mock_bucket


def _create_mock_type(filename="", content="", is_disable=False):
    """Returns a mock type

    Args:
        filename:
        content:
        is_disable:

    Returns:

    """
    mock_type = Mock(spec=Type)
    mock_type.filename = filename
    mock_type.content = content
    mock_type.id = ObjectId()
    mock_type.is_disabled = is_disable
    return mock_type


def _create_mock_type_version_manager(title="", versions=None, user="1"):
    """Returns a mock type version manager

    Args:
        title:
        versions:

    Returns:

    """
    mock_type_version_manager = Mock(spec=TypeVersionManager)
    mock_type_version_manager.title = title
    mock_type_version_manager.id = ObjectId()
    mock_type_version_manager.user = user
    if versions is not None:
        mock_type_version_manager.versions = versions
    else:
        mock_type_version_manager.versions = []
    mock_type_version_manager.disabled_versions = []
    return mock_type_version_manager


def _create_type(filename="", content=None):
    """Returns a type

    Args:
        filename:
        content:

    Returns:

    """
    if content is None:
        # set a valid content
        content = (
            "<schema xmlns='http://www.w3.org/2001/XMLSchema'><simpleType name='type'>"
            "<restriction base='string'><enumeration value='test'/></restriction>"
            "</simpleType></schema>"
        )
    return Type(id=ObjectId(), filename=filename, content=content)


def _create_type_version_manager(title="", user="1"):
    """Returns a type version manager

    Args:
        title:
        user:

    Returns:

    """
    return TypeVersionManager(
        id=ObjectId(), title=title, versions=[], user=user, disabled_versions=[]
    )



"""Serializers used throughout the bucket Rest API
"""
from django.http import Http404
from rest_framework.fields import CharField
from rest_framework_mongoengine.serializers import DocumentSerializer

from core_composer_app.components.bucket import api as bucket_api
from core_composer_app.components.bucket.models import Bucket
from core_main_app.commons.exceptions import DoesNotExist


class BucketSerializer(DocumentSerializer):
    """Bucket serializer"""

    class Meta(object):
        """Meta"""

        model = Bucket
        fields = "__all__"
        read_only_fields = (
            "id",
            "color",
        )

    def create(self, validated_data):
        """
        Create and return a new `Bucket` instance, given the validated data.
        """
        # Create data
        bucket = Bucket(
            label=validated_data["label"], types=validated_data.get("types", None)
        )
        # Save the data
        return bucket_api.upsert(bucket)

    def update(self, instance, validated_data):
        """
        Update and return an existing `Bucket` instance, given the validated data.
        """
        instance.label = validated_data.get("label", instance.label)
        return bucket_api.upsert(instance)


class BucketsSerializer(DocumentSerializer):
    """Buckets serializer."""

    id = CharField()

    class Meta(object):
        model = Bucket
        fields = ("id",)

    def validate_id(self, id):
        """Validate id field

        Args:
            id:

        Returns:

        """
        try:
            bucket_api.get_by_id(id)
            return id
        except DoesNotExist:
            raise Http404



""" Initialize permissions for core composer app.
"""
import logging

from django.contrib.auth.models import Permission, Group

from core_composer_app.permissions import rights as composer_rights
from core_main_app.permissions import rights as main_rights

logger = logging.getLogger(__name__)


def init_permissions():
    """Initialization of groups and permissions."""
    try:
        # Get or Create the default group
        default_group, created = Group.objects.get_or_create(
            name=main_rights.default_group
        )

        # Get composer permissions
        composer_access_perm = Permission.objects.get(
            codename=composer_rights.composer_access
        )
        composer_save_template_perm = Permission.objects.get(
            codename=composer_rights.composer_save_template
        )
        composer_save_type_perm = Permission.objects.get(
            codename=composer_rights.composer_save_type
        )

        # Add permissions to default group
        default_group.permissions.add(
            composer_access_perm, composer_save_template_perm, composer_save_type_perm
        )
    except Exception as e:
        logger.error("Impossible to init composer permissions: %s" % str(e))



"""Url router for the REST API
"""

from django.urls import re_path

from core_composer_app.rest.bucket import views as bucket_views
from core_composer_app.rest.type_version_manager import (
    views as type_version_manager_views,
)
from core_main_app.rest.template import views as template_views
from core_main_app.rest.template_version_manager import (
    views as template_version_manager_views,
)

urlpatterns = [
    re_path(
        r"^type-version-manager/global/$",
        type_version_manager_views.GlobalTypeVersionManagerList.as_view(),
        name="core_composer_app_rest_type_version_manager_global_list",
    ),
    re_path(
        r"^type-version-manager/user/$",
        type_version_manager_views.UserTypeVersionManagerList.as_view(),
        name="core_composer_app_rest_type_version_manager_user_list",
    ),
    re_path(
        r"^type-version-manager/(?P<pk>\w+)/$",
        template_version_manager_views.TemplateVersionManagerDetail.as_view(),
        name="core_composer_app_rest_type_version_manager_detail",
    ),
    re_path(
        r"^type-version-manager/(?P<pk>\w+)/disable/$",
        template_version_manager_views.DisableTemplateVersionManager.as_view(),
        name="core_composer_app_rest_type_version_manager_disable",
    ),
    re_path(
        r"^type-version-manager/(?P<pk>\w+)/restore/$",
        template_version_manager_views.RestoreTemplateVersionManager.as_view(),
        name="core_composer_app_rest_type_version_manager_restore",
    ),
    re_path(
        r"^type/version/(?P<pk>\w+)/current/$",
        template_version_manager_views.CurrentTemplateVersion.as_view(),
        name="core_composer_app_rest_type_version_current",
    ),
    re_path(
        r"^type/version/(?P<pk>\w+)/disable/$",
        template_version_manager_views.DisableTemplateVersion.as_view(),
        name="core_composer_app_rest_type_version_disable",
    ),
    re_path(
        r"^type/global/$",
        type_version_manager_views.GlobalTypeList.as_view(),
        name="core_composer_app_rest_global_type_list",
    ),
    re_path(
        r"^type/$",
        type_version_manager_views.UserTypeList.as_view(),
        name="core_composer_app_rest_user_type_list",
    ),
    re_path(
        r"^type/version/(?P<pk>\w+)/restore/$",
        template_version_manager_views.RestoreTemplateVersion.as_view(),
        name="core_composer_app_rest_type_version_restore",
    ),
    re_path(
        r"^type/(?P<pk>\w+)/download/$",
        template_views.TemplateDownload.as_view(),
        name="core_composer_app_rest_type_download",
    ),
    re_path(
        r"^type/(?P<pk>\w+)/$",
        template_views.TemplateDetail.as_view(),
        name="core_composer_app_rest_type_detail",
    ),
    re_path(
        r"^bucket/$",
        bucket_views.BucketList.as_view(),
        name="core_composer_app_rest_bucket_list",
    ),
    re_path(
        r"^bucket/(?P<pk>\w+)/$",
        bucket_views.BucketDetail.as_view(),
        name="core_composer_app_rest_bucket_detail",
    ),
    re_path(
        r"^buckets/type-version-manager/(?P<pk>\w+)/$",
        bucket_views.TypeVersionManagerBuckets.as_view(),
        name="core_composer_app_rest_type_version_manger_buckets",
    ),
]



""" Fixtures files for type version manager
"""
from core_composer_app.components.bucket.models import Bucket
from tests.components.type_version_manager.fixtures.fixtures import (
    TypeVersionManagerFixtures,
)


class BucketFixtures(TypeVersionManagerFixtures):
    """Bucket fixtures"""

    bucket_empty = None
    bucket_1 = None
    bucket_2 = None
    bucket_collection = None

    def insert_data(self):
        """Insert a set of Buckets.

        Returns:

        """
        super(BucketFixtures, self).insert_data()

        self.bucket_empty = Bucket(label="empty", color="#000000", types=[]).save()
        self.bucket_1 = Bucket(
            label="bucket1", color="#000001", types=[self.type_vm_1]
        ).save()
        self.bucket_2 = Bucket(
            label="bucket2", color="#000002", types=[self.type_vm_1, self.type_vm_2]
        ).save()

        self.bucket_collection = [self.bucket_empty, self.bucket_1, self.bucket_2]



""" Authentication tests for Bucket REST API
"""
from django.test import SimpleTestCase
from mock.mock import patch
from rest_framework import status

from core_composer_app.components.bucket.models import Bucket
from core_composer_app.rest.bucket import views as bucket_views
from core_composer_app.rest.bucket.serializers import BucketSerializer
from core_main_app.components.version_manager.models import VersionManager
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from core_main_app.utils.tests_tools.RequestMock import RequestMock


class TestBucketListGetPermission(SimpleTestCase):
    def test_anonymous_returns_http_403(self):
        response = RequestMock.do_request_get(bucket_views.BucketList.as_view(), None)

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch.object(Bucket, "get_all")
    def test_authenticated_returns_http_200(self, bucket_get_all):
        bucket_get_all.return_value = {}

        mock_user = create_mock_user("1")

        response = RequestMock.do_request_get(
            bucket_views.BucketList.as_view(), mock_user
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(Bucket, "get_all")
    def test_staff_returns_http_200(self, bucket_get_all):
        bucket_get_all.return_value = {}

        mock_user = create_mock_user("1", is_staff=True)

        response = RequestMock.do_request_get(
            bucket_views.BucketList.as_view(), mock_user
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)


class TestBucketListPostPermission(SimpleTestCase):
    def test_anonymous_returns_http_403(self):
        response = RequestMock.do_request_post(bucket_views.BucketList.as_view(), None)

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_authenticated_returns_http_403(self):
        mock_user = create_mock_user("1")

        response = RequestMock.do_request_post(
            bucket_views.BucketList.as_view(), mock_user
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch.object(BucketSerializer, "is_valid")
    @patch.object(BucketSerializer, "save")
    @patch.object(BucketSerializer, "data")
    def test_staff_returns_http_201(
        self, bucket_serializer_data, bucket_serializer_save, bucket_serializer_valid
    ):
        bucket_serializer_data.return_value = True
        bucket_serializer_save.return_value = None
        bucket_serializer_valid.return_value = {}

        mock_user = create_mock_user("1", is_staff=True)

        response = RequestMock.do_request_post(
            bucket_views.BucketList.as_view(), mock_user
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)


class TestBucketDetailGetPermission(SimpleTestCase):
    def setUp(self):
        self.fake_id = "507f1f77bcf86cd799439011"

    def test_anonymous_returns_http_403(self):
        response = RequestMock.do_request_get(
            bucket_views.BucketDetail.as_view(), None, param={"pk": self.fake_id}
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch.object(Bucket, "get_by_id")
    @patch.object(BucketSerializer, "data")
    def test_authenticated_returns_http_200(
        self, bucket_get_by_id, bucket_serializer_data
    ):
        bucket_get_by_id.return_value = {}
        bucket_serializer_data.return_value = True

        mock_user = create_mock_user("1")

        response = RequestMock.do_request_get(
            bucket_views.BucketDetail.as_view(), mock_user, param={"pk": self.fake_id}
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(Bucket, "get_by_id")
    @patch.object(BucketSerializer, "data")
    def test_staff_returns_http_200(self, bucket_get_by_id, bucket_serializer_data):
        bucket_get_by_id.return_value = {}
        bucket_serializer_data.return_value = True

        mock_user = create_mock_user("1", is_staff=True)

        response = RequestMock.do_request_get(
            bucket_views.BucketDetail.as_view(), mock_user, param={"pk": self.fake_id}
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)


class TestBucketDetailDeletePermission(SimpleTestCase):
    def setUp(self):
        self.fake_id = "507f1f77bcf86cd799439011"

    def test_anonymous_returns_http_403(self):
        response = RequestMock.do_request_delete(
            bucket_views.BucketDetail.as_view(), None, param={"pk": self.fake_id}
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_authenticated_returns_http_403(self):
        mock_user = create_mock_user("1")

        response = RequestMock.do_request_delete(
            bucket_views.BucketDetail.as_view(), mock_user, param={"pk": self.fake_id}
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch.object(Bucket, "get_by_id")
    @patch("core_composer_app.components.bucket.api.delete")
    def test_staff_returns_http_204(self, bucket_get_by_id, bucket_api_delete):
        bucket_get_by_id.return_value = {}
        bucket_api_delete.return_value = None

        mock_user = create_mock_user("1", is_staff=True)

        response = RequestMock.do_request_delete(
            bucket_views.BucketDetail.as_view(), mock_user, param={"pk": self.fake_id}
        )

        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)


class TestBucketDetailPatchPermission(SimpleTestCase):
    def setUp(self):
        self.fake_id = "507f1f77bcf86cd799439011"

    def test_anonymous_returns_http_403(self):

        response = RequestMock.do_request_patch(
            bucket_views.BucketDetail.as_view(), None, param={"pk": self.fake_id}
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_authenticated_returns_http_403(self):
        mock_user = create_mock_user("1")

        response = RequestMock.do_request_patch(
            bucket_views.BucketDetail.as_view(), mock_user, param={"pk": self.fake_id}
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch.object(Bucket, "get_by_id")
    @patch.object(BucketSerializer, "is_valid")
    @patch.object(BucketSerializer, "save")
    @patch.object(BucketSerializer, "data")
    def test_staff_returns_http_200(
        self,
        bucket_serializer_data,
        bucket_serializer_save,
        bucket_serializer_valid,
        bucket_get_by_id,
    ):
        bucket_get_by_id.return_value = {}
        bucket_serializer_data.return_value = True
        bucket_serializer_save.return_value = None
        bucket_serializer_valid.return_value = {}

        mock_user = create_mock_user("1", is_staff=True)

        response = RequestMock.do_request_patch(
            bucket_views.BucketDetail.as_view(), mock_user, param={"pk": self.fake_id}
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)


class TestTypeVersionManagerBucketsPatchPermission(SimpleTestCase):
    def setUp(self):
        self.fake_id = "507f1f77bcf86cd799439011"

    def test_anonymous_returns_http_403(self):
        response = RequestMock.do_request_patch(
            bucket_views.TypeVersionManagerBuckets.as_view(),
            None,
            param={"pk": self.fake_id},
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_authenticated_returns_http_403(self):
        mock_user = create_mock_user("1")

        response = RequestMock.do_request_patch(
            bucket_views.TypeVersionManagerBuckets.as_view(),
            mock_user,
            param={"pk": self.fake_id},
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch.object(Bucket, "get_by_id")
    @patch.object(VersionManager, "get_by_id")
    @patch.object(BucketSerializer, "is_valid")
    @patch.object(BucketSerializer, "save")
    @patch.object(BucketSerializer, "data")
    def test_staff_returns_http_200(
        self,
        bucket_serializer_data,
        bucket_serializer_save,
        bucket_serializer_valid,
        version_manager_get_by_id,
        bucket_get_by_id,
    ):
        bucket_get_by_id.return_value = {}
        version_manager_get_by_id.return_value = VersionManager(user=None)
        bucket_serializer_data.return_value = True
        bucket_serializer_save.return_value = None
        bucket_serializer_valid.return_value = {}
        mock_user = create_mock_user("1", is_staff=True)

        response = RequestMock.do_request_patch(
            bucket_views.TypeVersionManagerBuckets.as_view(),
            mock_user,
            param={"pk": self.fake_id},
            data=[],
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)



""" Url router for the composer application
"""
from django.conf import settings
from django.conf.urls import include
from django.urls import re_path

from core_composer_app.views.user import views as user_views, ajax as user_ajax

parser_url = []
if "core_parser_app" in settings.INSTALLED_APPS:
    from core_parser_app.views.common import views as common_parser_views

    parser_url = [
        re_path(
            r"^type/modules/(?P<pk>\w+)",
            common_parser_views.ManageModulesUserView.as_view(
                back_to_previous_url="core_composer_app_manage_type_versions"
            ),
            name="core_composer_app_type_modules",
        ),
    ]

urlpatterns = [
    re_path(r"^$", user_views.index, name="core_composer_index"),
    re_path(
        r"^build-template/(?P<template_id>\w+)$",
        user_views.build_template,
        name="core_composer_build_template",
    ),
    re_path(
        r"^download-xsd$", user_views.download_xsd, name="core_composer_download_xsd"
    ),
    re_path(
        r"^type/versions/(?P<version_manager_id>\w+)",
        user_views.manage_type_versions,
        name="core_composer_app_manage_type_versions",
    ),
    re_path(
        r"^change-xsd-type$",
        user_ajax.change_xsd_type,
        name="core_composer_change_xsd_type",
    ),
    re_path(
        r"^change-root-type-name$",
        user_ajax.change_root_type_name,
        name="core_composer_change_root_type_name",
    ),
    re_path(
        r"^insert-element-sequence$",
        user_ajax.insert_element_sequence,
        name="core_composer_insert_element_sequence",
    ),
    re_path(
        r"^rename-element$",
        user_ajax.rename_element,
        name="core_composer_rename_element",
    ),
    re_path(
        r"^delete-element$",
        user_ajax.delete_element,
        name="core_composer_delete_element",
    ),
    re_path(
        r"^get-element-occurrences$",
        user_ajax.get_element_occurrences,
        name="core_composer_get_element_occurrences",
    ),
    re_path(
        r"^set-element-occurrences$",
        user_ajax.set_element_occurrences,
        name="core_composer_set_element_occurrences",
    ),
    re_path(
        r"^save-template$", user_ajax.save_template, name="core_composer_save_template"
    ),
    re_path(r"^save-type$", user_ajax.save_type, name="core_composer_save_type"),
    re_path(r"^rest/", include("core_composer_app.rest.urls")),
]

urlpatterns = urlpatterns + parser_url



""" Add Composer in main menu
"""
from django.urls import reverse
from menu import Menu, MenuItem

Menu.add_item("main", MenuItem("Composer", reverse("core_composer_index")))


types_children = (
    MenuItem("Type List", reverse("admin:core_composer_app_types"), icon="list"),
    MenuItem(
        "Upload New Type", reverse("admin:core_composer_app_upload_type"), icon="upload"
    ),
    MenuItem("Manage Buckets", reverse("admin:core_composer_app_buckets"), icon="tags"),
)

Menu.add_item("admin", MenuItem("TYPES", None, children=types_children))



"""Url router for the administration site
"""
from django.conf import settings
from django.contrib import admin
from django.contrib.admin.views.decorators import staff_member_required
from django.urls import re_path, reverse_lazy

from core_composer_app.views.admin import views as admin_views, ajax as admin_ajax
from core_composer_app.views.admin.ajax import EditBucketView
from core_main_app.views.common.ajax import EditTemplateVersionManagerView

parser_url = []
if "core_parser_app" in settings.INSTALLED_APPS:
    from core_parser_app.views.admin import views as admin_parser_views

    parser_url = [
        re_path(
            r"^type/modules/(?P<pk>\w+)",
            staff_member_required(
                admin_parser_views.ManageModulesAdminView.as_view(
                    back_to_previous_url="admin:core_composer_app_manage_type_versions"
                )
            ),
            name="core_composer_app_type_modules",
        ),
    ]

admin_urls = [
    re_path(r"^types$", admin_views.manage_types, name="core_composer_app_types"),
    re_path(
        r"^type/(?P<pk>[\w-]+)/edit/$",
        staff_member_required(
            EditTemplateVersionManagerView.as_view(
                success_url=reverse_lazy("admin:core_composer_app_types")
            )
        ),
        name="core_composer_app_edit_type",
    ),
    re_path(
        r"^type/upload$", admin_views.upload_type, name="core_composer_app_upload_type"
    ),
    re_path(
        r"^type/upload/(?P<version_manager_id>\w+)",
        admin_views.upload_type_version,
        name="core_composer_app_upload_type_version",
    ),
    re_path(
        r"^type/versions/(?P<version_manager_id>\w+)",
        admin_views.manage_type_versions,
        name="core_composer_app_manage_type_versions",
    ),
    re_path(
        r"^type/buckets/(?P<version_manager_id>\w+)",
        admin_views.manage_type_buckets,
        name="core_composer_app_type_buckets",
    ),
    re_path(r"^buckets$", admin_views.manage_buckets, name="core_composer_app_buckets"),
    re_path(
        r"^bucket/upload$",
        admin_views.upload_bucket,
        name="core_composer_app_upload_bucket",
    ),
    re_path(
        r"^bucket/delete$",
        admin_ajax.delete_bucket,
        name="core_composer_app_delete_bucket",
    ),
    re_path(
        r"^type/resolve-dependencies",
        admin_ajax.resolve_dependencies,
        name="core_composer_app_resolve_dependencies",
    ),
    re_path(
        r"^bucket/(?P<pk>[\w-]+)/edit/$",
        staff_member_required(EditBucketView.as_view()),
        name="core_composer_app_edit_bucket",
    ),
]

urls = admin.site.get_urls()
admin.site.get_urls = lambda: admin_urls + urls + parser_url



"""Unit tests for composer xml operations
"""
# TODO: add tests for other xml utils

from os.path import join, dirname, abspath
from unittest.case import TestCase

from core_composer_app.utils.xml import (
    _insert_element_type,
    check_type_core_support,
    COMPLEX_TYPE,
    SIMPLE_TYPE,
)
from core_main_app.commons.exceptions import CoreError
from core_main_app.utils.xml import validate_xml_schema

RESOURCES_PATH = join(dirname(abspath(__file__)), "data")


class TestInsertElementType(TestCase):
    def setUp(self):
        self.root_xpath = "xsd:complexType/xsd:sequence"
        self.type_name = "new"

    def test_add_type_to_base(self):
        base_filename = "base.xsd"
        type_filename = "type.xsd"
        # load test resources
        with open(join(RESOURCES_PATH, base_filename), "r") as base_file:
            base_content = base_file.read()
        with open(join(RESOURCES_PATH, type_filename), "r") as type_file:
            type_content = type_file.read()

        result_tree = _insert_element_type(
            base_content,
            self.root_xpath,
            type_content,
            self.type_name,
            join(RESOURCES_PATH, type_filename),
        )

        errors = validate_xml_schema(result_tree, request=None)
        self.assertTrue(errors is None)

    def test_add_type_with_target_namespace_to_base(self):
        base_filename = "base.xsd"
        type_filename = "type_target_ns.xsd"
        # load test resources
        with open(join(RESOURCES_PATH, base_filename), "r") as base_file:
            base_content = base_file.read()
        with open(join(RESOURCES_PATH, type_filename), "r") as type_file:
            type_content = type_file.read()

        with self.assertRaises(CoreError):
            _insert_element_type(
                base_content,
                self.root_xpath,
                type_content,
                self.type_name,
                join(RESOURCES_PATH, type_filename),
            )

    def test_add_type_to_base_with_target_namespace(self):
        base_filename = "base_target_ns.xsd"
        type_filename = "type.xsd"
        # load test resources
        with open(join(RESOURCES_PATH, base_filename), "r") as base_file:
            base_content = base_file.read()
        with open(join(RESOURCES_PATH, type_filename), "r") as type_file:
            type_content = type_file.read()

        result_tree = _insert_element_type(
            base_content,
            self.root_xpath,
            type_content,
            self.type_name,
            join(RESOURCES_PATH, type_filename),
        )

        errors = validate_xml_schema(result_tree, request=None)
        self.assertTrue(errors is None)

    def test_add_type_with_target_namespace_to_base_with_same_target_namespace(self):
        base_filename = "base_target_ns.xsd"
        type_filename = "type_base_target_ns.xsd"
        # load test resources
        with open(join(RESOURCES_PATH, base_filename), "r") as base_file:
            base_content = base_file.read()
        with open(join(RESOURCES_PATH, type_filename), "r") as type_file:
            type_content = type_file.read()

        result_tree = _insert_element_type(
            base_content,
            self.root_xpath,
            type_content,
            self.type_name,
            join(RESOURCES_PATH, type_filename),
        )

        errors = validate_xml_schema(result_tree, request=None)
        self.assertTrue(errors is None)

    def test_add_type_with_target_namespace_to_base_with_different_target_namespace(
        self,
    ):
        base_filename = "base_target_ns.xsd"
        type_filename = "type_target_ns.xsd"
        # load test resources
        with open(join(RESOURCES_PATH, base_filename), "r") as base_file:
            base_content = base_file.read()
        with open(join(RESOURCES_PATH, type_filename), "r") as type_file:
            type_content = type_file.read()

        with self.assertRaises(CoreError):
            _insert_element_type(
                base_content,
                self.root_xpath,
                type_content,
                self.type_name,
                join(RESOURCES_PATH, type_filename),
            )

    def test_add_type_with_target_namespace_prefix_to_base(self):
        base_filename = "base.xsd"
        type_filename = "type_target_ns_prefix.xsd"
        # load test resources
        with open(join(RESOURCES_PATH, base_filename), "r") as base_file:
            base_content = base_file.read()
        with open(join(RESOURCES_PATH, type_filename), "r") as type_file:
            type_content = type_file.read()

        result_tree = _insert_element_type(
            base_content,
            self.root_xpath,
            type_content,
            self.type_name,
            join(RESOURCES_PATH, type_filename),
        )

        errors = validate_xml_schema(result_tree, request=None)
        self.assertTrue(errors is None)

    def test_add_type_with_target_namespace_prefix_to_base_with_same_target_namespace_prefix(
        self,
    ):
        base_filename = "base_target_ns.xsd"
        type_filename = "type_base_target_ns_prefix.xsd"
        # load test resources
        with open(join(RESOURCES_PATH, base_filename), "r") as base_file:
            base_content = base_file.read()
        with open(join(RESOURCES_PATH, type_filename), "r") as type_file:
            type_content = type_file.read()

        result_tree = _insert_element_type(
            base_content,
            self.root_xpath,
            type_content,
            self.type_name,
            join(RESOURCES_PATH, type_filename),
        )

        errors = validate_xml_schema(result_tree, request=None)
        self.assertTrue(errors is None)

    def test_add_type_with_target_namespace_prefix_to_base_with_different_target_namespace_prefix(
        self,
    ):
        base_filename = "base_target_ns.xsd"
        type_filename = "type_target_ns_prefix.xsd"
        # load test resources
        with open(join(RESOURCES_PATH, base_filename), "r") as base_file:
            base_content = base_file.read()
        with open(join(RESOURCES_PATH, type_filename), "r") as type_file:
            type_content = type_file.read()

        result_tree = _insert_element_type(
            base_content,
            self.root_xpath,
            type_content,
            self.type_name,
            join(RESOURCES_PATH, type_filename),
        )

        errors = validate_xml_schema(result_tree, request=None)
        self.assertTrue(errors is None)


class TestTypeDefinition(TestCase):
    def test_simple_type(self):
        type_filename = "type.xsd"
        # load test resources
        with open(join(RESOURCES_PATH, type_filename), "r") as type_file:
            type_content = type_file.read()

        type_content = check_type_core_support(type_content)

        self.assertEqual(type_content, SIMPLE_TYPE)

    def test_complex_type(self):
        type_filename = "type_complex.xsd"
        # load test resources
        with open(join(RESOURCES_PATH, type_filename), "r") as type_file:
            type_content = type_file.read()

        type_content = check_type_core_support(type_content)

        self.assertEqual(type_content, COMPLEX_TYPE)



"""
Type Version Manager API
"""
from core_composer_app.components.bucket import api as bucket_api
from core_composer_app.components.type import api as type_api
from core_composer_app.components.type_version_manager.models import TypeVersionManager
from core_main_app.access_control.api import is_superuser
from core_main_app.access_control.decorators import access_control
from core_main_app.components.template import api as template_api
from core_main_app.components.template.access_control import can_read_global
from core_main_app.components.template_version_manager.access_control import can_write
from core_main_app.components.version_manager import api as version_manager_api
from core_main_app.components.version_manager.utils import get_latest_version_name


@access_control(can_write)
def insert(type_version_manager, type_object, request, list_bucket_ids=None):
    """Add a version to a type version manager.

    Args:
        type_version_manager:
        type_object:
        request:
        list_bucket_ids:

    Returns:

    """
    # save the type in database
    type_api.upsert(type_object, request=request)
    try:
        # insert the initial template in the version manager
        version_manager_api.insert_version(
            type_version_manager, type_object, request=request
        )
        # insert the version manager in database
        version_manager_api.upsert(type_version_manager, request=request)
        # Add to the selected buckets
        bucket_api.add_type_to_buckets(type_version_manager, list_bucket_ids)
        # get type display name
        display_name = get_latest_version_name(type_version_manager)
        # update saved template
        template_api.set_display_name(type_object, display_name, request=request)
        # update saved type
        return type_version_manager
    except Exception as e:
        type_api.delete(type_object, request=request)
        raise e


@access_control(can_read_global)
def get_global_version_managers(request):
    """Get all global version managers of a type.

    Args:
        request:

    Returns:

    """
    return TypeVersionManager.get_global_version_managers()


@access_control(can_read_global)
def get_active_global_version_manager(request):
    """Return all active Version Managers with user set to None.

    Args:
        request:

    Returns:

    """
    return TypeVersionManager.get_active_global_version_manager()


def get_version_managers_by_user(request):
    """Get all global version managers of a user.

    Returns:

    """
    return TypeVersionManager.get_version_managers_by_user(str(request.user.id))


def get_active_version_manager_by_user_id(request):
    """Return all active Version Managers with given user id.

    Returns:

    """
    return TypeVersionManager.get_active_version_manager_by_user_id(
        str(request.user.id)
    )


@access_control(can_read_global)
def get_no_buckets_types(request):
    """Get list of available types not inside a bucket.

    Returns:

    """
    # build list of types
    bucket_types = []
    for bucket in bucket_api.get_all():
        bucket_types += bucket.types

    all_types = get_global_version_managers(request=request)
    no_bucket_types = [
        type_version_manager
        for type_version_manager in all_types
        if type_version_manager not in bucket_types
    ]
    return no_bucket_types


@access_control(is_superuser)
def get_all_version_manager(request):
    """Return all Type Version Managers of all users.

    Returns:

    """
    return TypeVersionManager.get_all_type_version_manager()



"""Composer Admin Forms
"""
from django import forms

from core_composer_app.components.bucket import api as bucket_api
from core_composer_app.components.bucket.models import Bucket
from django_mongoengine.forms import DocumentForm


class BucketForm(forms.Form):
    """
    Form to add a bucket.
    """

    label = forms.CharField(
        label="Enter Bucket label",
        max_length=100,
        required=True,
        widget=forms.TextInput(attrs={"class": "form-control"}),
    )


class BucketDataModelChoiceField(forms.ModelMultipleChoiceField):
    """
    Choice Field to select an existing form.
    """

    def label_from_instance(self, obj):
        return obj.label


class UploadTypeForm(forms.Form):
    """
    Form to upload a new Type.
    """

    name = forms.CharField(
        label="Enter Type name",
        max_length=100,
        required=True,
        widget=forms.TextInput(attrs={"class": "form-control"}),
    )
    xsd_file = forms.FileField(
        label="Select a file",
        required=True,
        widget=forms.FileInput(attrs={"accept": ".xsd", "class": "form-control"}),
    )
    buckets = BucketDataModelChoiceField(
        label="Select buckets",
        queryset=bucket_api.get_all(),
        required=False,
        widget=forms.SelectMultiple(attrs={"class": "form-control"}),
    )


class EditTypeBucketsForm(forms.Form):
    """
    Form to edit buckets of a Type.
    """

    buckets = BucketDataModelChoiceField(
        label="Select new buckets",
        queryset=bucket_api.get_all(),
        required=False,
        widget=forms.SelectMultiple(attrs={"class": "form-control"}),
    )


class EditBucketForm(DocumentForm):
    label = forms.CharField(
        label="Label",
        widget=forms.TextInput(
            attrs={"class": "form-control", "placeholder": "Enter the new label"}
        ),
    )

    class Meta(object):
        document = Bucket
        fields = ["label"]



""" Integration Test for Type Version Manager Rest API
"""
import json

from django.urls import reverse
from django.test import override_settings
from rest_framework import status

from core_composer_app.components.type import api as type_api
from core_composer_app.rest.type_version_manager import views
from core_main_app.components.version_manager import api as vm_api
from core_main_app.utils.integration_tests.integration_base_test_case import (
    MongoIntegrationBaseTestCase,
)
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from core_main_app.utils.tests_tools.RequestMock import RequestMock, create_mock_request
from tests.components.type_version_manager.fixtures.fixtures import (
    TypeVersionManagerFixtures,
)

fixture_type = TypeVersionManagerFixtures()


class TestGlobalTypeVersionManagerList(MongoIntegrationBaseTestCase):
    fixture = fixture_type

    def setUp(self):
        super(TestGlobalTypeVersionManagerList, self).setUp()

    def test_get_returns_http_200(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            views.GlobalTypeVersionManagerList.as_view(), user
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_get_returns_all_global_tvm(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            views.GlobalTypeVersionManagerList.as_view(), user
        )

        # Assert
        self.assertEqual(len(response.data), 1)

    def test_get_returned_tvm_are_global(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            views.GlobalTypeVersionManagerList.as_view(), user
        )

        # Assert
        self.assertEqual(response.data[0]["user"], None)

    def test_get_filtered_by_correct_title_returns_tvm(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            views.GlobalTypeVersionManagerList.as_view(),
            user,
            data={"title": self.fixture.type_vm_1.title},
        )

        # Assert
        self.assertEqual(len(response.data), 1)

    def test_get_filtered_by_incorrect_title_returns_no_tvm(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            views.GlobalTypeVersionManagerList.as_view(),
            user,
            data={"title": "bad title"},
        )

        # Assert
        self.assertEqual(len(response.data), 0)

    def test_get_filtered_by_expected_is_disabled_returns_tvm(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            views.GlobalTypeVersionManagerList.as_view(),
            user,
            data={"is_disabled": self.fixture.type_vm_1.is_disabled},
        )

        # Assert
        self.assertEqual(len(response.data), 1)

    def test_get_filtered_by_incorrect_is_disabled_returns_no_tvm(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            views.GlobalTypeVersionManagerList.as_view(),
            user,
            data={"is_disabled": not self.fixture.type_vm_1.is_disabled},
        )

        # Assert
        self.assertEqual(len(response.data), 0)


class TestUserTypeVersionManagerList(MongoIntegrationBaseTestCase):
    fixture = fixture_type

    def setUp(self):
        super(TestUserTypeVersionManagerList, self).setUp()

    def test_get_returns_http_200(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            views.UserTypeVersionManagerList.as_view(), user
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_get_returns_all_user_tvm(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            views.UserTypeVersionManagerList.as_view(), user
        )

        # Assert
        self.assertEqual(len(response.data), 1)

    def test_get_returned_tvm_are_from_user(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            views.UserTypeVersionManagerList.as_view(), user
        )

        # Assert
        self.assertEqual(response.data[0]["user"], "1")

    def test_get_filtered_by_correct_title_returns_tvm(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            views.UserTypeVersionManagerList.as_view(),
            user,
            data={"title": self.fixture.type_vm_2.title},
        )

        # Assert
        self.assertEqual(len(response.data), 1)

    def test_get_filtered_by_incorrect_title_returns_no_tvm(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            views.UserTypeVersionManagerList.as_view(),
            user,
            data={"title": "bad title"},
        )

        # Assert
        self.assertEqual(len(response.data), 0)

    def test_get_filtered_by_expected_is_disabled_returns_tvm(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            views.UserTypeVersionManagerList.as_view(),
            user,
            data={"is_disabled": self.fixture.type_vm_2.is_disabled},
        )

        # Assert
        self.assertEqual(len(response.data), 1)

    def test_get_filtered_by_incorrect_is_disabled_returns_no_tvm(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_get(
            views.UserTypeVersionManagerList.as_view(),
            user,
            data={"is_disabled": not self.fixture.type_vm_2.is_disabled},
        )

        # Assert
        self.assertEqual(len(response.data), 0)


class TestUserTypeList(MongoIntegrationBaseTestCase):
    fixture = fixture_type

    def setUp(self):
        super(TestUserTypeList, self).setUp()
        type_content = (
            "<xs:schema xmlns:xs='http://www.w3.org/2001/XMLSchema'>"
            "<xs:simpleType name='temperature'>"
            "<xs:restriction base='xs:string'>"
            "<xs:enumeration value='Kelvin'/>"
            "<xs:enumeration value='Celsius'/>"
            "<xs:enumeration value='Fahrenheit'/>"
            "</xs:restriction></xs:simpleType>"
            "</xs:schema>"
        )
        self.data = {"title": "title", "filename": "filename", "content": type_content}

    @override_settings(ROOT_URLCONF="core_main_app.urls")
    def test_post_returns_http_201(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_post(
            views.UserTypeList.as_view(), user, data=self.data
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    @override_settings(ROOT_URLCONF="core_main_app.urls")
    def test_post_owner_is_user(self):
        # Arrange
        user = create_mock_user("1")
        mock_request = create_mock_request(user=user)

        # Act
        response = RequestMock.do_request_post(
            views.UserTypeList.as_view(), user, data=self.data
        )

        # get type version manager from posted type
        type_id = response.data["id"]
        type_object = type_api.get(type_id, request=mock_request)
        type_version_manager = vm_api.get_from_version(
            type_object, request=mock_request
        )

        # Assert
        self.assertEqual(type_version_manager.user, user.id)

    def test_post_type_name_already_exists_returns_http_400(self):
        # Arrange
        user = create_mock_user("1")
        self.data["title"] = self.fixture.type_vm_1.title

        # Act
        response = RequestMock.do_request_post(
            views.UserTypeList.as_view(), user, data=self.data
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    @override_settings(ROOT_URLCONF="core_main_app.urls")
    def test_post_type_with_correct_dependency_returns_http_201(self):
        # Arrange
        user = create_mock_user("1")
        self.data["content"] = (
            "<xs:schema xmlns:xs='http://www.w3.org/2001/XMLSchema'>"
            "<xs:include schemaLocation='type1_1.xsd'/>"
            "<xs:simpleType name='root'/></xs:schema>"
        )

        self.data["dependencies_dict"] = json.dumps(
            {"type1_1.xsd": str(self.fixture.type_1_1.id)}
        )

        # Act
        response = RequestMock.do_request_post(
            views.UserTypeList.as_view(), user, data=self.data
        )

        # Assert
        # FIXME: unable to download self dependency because server not running
        # self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        expected_download_url = reverse(
            "core_main_app_rest_template_download",
            kwargs={"pk": self.fixture.type_1_1.id},
        )
        self.assertTrue(expected_download_url in response.data["message"])

    @override_settings(ROOT_URLCONF="core_main_app.urls")
    def test_post_type_with_incorrect_dependency_schemaLocation_returns_http_400(self):
        # Arrange
        user = create_mock_user("1")
        self.data["content"] = (
            "<xs:schema xmlns:xs='http://www.w3.org/2001/XMLSchema'>"
            "<xs:include schemaLocation='type1_1.xsd'/>"
            "<xs:simpleType name='root'/></xs:schema>"
        )

        self.data["dependencies_dict"] = json.dumps(
            {"test.xsd": str(self.fixture.type_1_1.id)}
        )

        # Act
        response = RequestMock.do_request_post(
            views.UserTypeList.as_view(), user, data=self.data
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    @override_settings(ROOT_URLCONF="core_main_app.urls")
    def test_post_type_with_incorrect_dependency_id_returns_http_400(self):
        # Arrange
        user = create_mock_user("1")
        self.data["content"] = (
            "<xs:schema xmlns:xs='http://www.w3.org/2001/XMLSchema'>"
            "<xs:include schemaLocation='type1_1.xsd'/>"
            "<xs:simpleType name='root'/></xs:schema>"
        )

        self.data["dependencies_dict"] = json.dumps({"type1_1.xsd": "bad_id"})

        # Act
        response = RequestMock.do_request_post(
            views.UserTypeList.as_view(), user, data=self.data
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)


class TestGlobalTypeList(MongoIntegrationBaseTestCase):
    fixture = fixture_type

    def setUp(self):
        super(TestGlobalTypeList, self).setUp()
        type_content = (
            "<xs:schema xmlns:xs='http://www.w3.org/2001/XMLSchema'>"
            "<xs:simpleType name='temperature'>"
            "<xs:restriction base='xs:string'>"
            "<xs:enumeration value='Kelvin'/>"
            "<xs:enumeration value='Celsius'/>"
            "<xs:enumeration value='Fahrenheit'/>"
            "</xs:restriction></xs:simpleType>"
            "</xs:schema>"
        )
        self.data = {"title": "title", "filename": "filename", "content": type_content}

    @override_settings(ROOT_URLCONF="core_main_app.urls")
    def test_post_returns_http_201_if_user_is_staff(self):
        # Arrange
        user = create_mock_user("1", is_staff=True)

        # Act
        response = RequestMock.do_request_post(
            views.GlobalTypeList.as_view(), user, data=self.data
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    @override_settings(ROOT_URLCONF="core_main_app.urls")
    def test_post_returns_http_403_if_user_is_superuser(self):
        # Arrange
        user = create_mock_user("1", is_superuser=True)

        # Act
        response = RequestMock.do_request_post(
            views.GlobalTypeList.as_view(), user, data=self.data
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @override_settings(ROOT_URLCONF="core_main_app.urls")
    def test_post_returns_http_403_if_user_does_not_have_permission(self):
        # Arrange
        user = create_mock_user("1")

        # Act
        response = RequestMock.do_request_post(
            views.GlobalTypeList.as_view(), user, data=self.data
        )

        # Assert
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @override_settings(ROOT_URLCONF="core_main_app.urls")
    def test_post_owner_is_global(self):
        # Arrange
        user = create_mock_user("1", is_staff=True)
        mock_request = create_mock_request(user=user)

        # Act
        response = RequestMock.do_request_post(
            views.GlobalTypeList.as_view(), user, data=self.data
        )

        # get type version manager from posted type
        type_id = response.data["id"]
        type_object = type_api.get(type_id, request=mock_request)
        type_version_manager = vm_api.get_from_version(
            type_object, request=mock_request
        )

        # Assert
        self.assertEqual(type_version_manager.user, None)



""" Views for the Bucket REST API
"""
from django.http import Http404
from django.utils.decorators import method_decorator
from rest_framework import status
from rest_framework.exceptions import ValidationError
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView

from core_composer_app.components.bucket import api as bucket_api
from core_composer_app.rest.bucket.serializers import (
    BucketSerializer,
    BucketsSerializer,
)
from core_main_app.commons.exceptions import DoesNotExist
from core_main_app.rest.template_version_manager.abstract_views import (
    AbstractTemplateVersionManagerDetail,
)
from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.utils.decorators import api_staff_member_required


class BucketList(APIView):
    """List all buckets, or create a new one."""

    permission_classes = (IsAuthenticated,)

    def get(self, request):
        """Get all buckets

        Url Parameters:

            label: label

        Examples:

            ../bucket/
            ../bucket?label=[label]

        Args:

            request: HTTP request

        Returns:

            - code: 200
              content: List of buckets
            - code: 500
              content: Internal server error
        """
        try:
            # Get objects
            object_list = bucket_api.get_all()

            # Apply filters
            label = self.request.query_params.get("label", None)
            if label is not None:
                object_list = object_list.filter(label=label)

            # Serialize object
            serializer = BucketSerializer(object_list, many=True)
            return Response(serializer.data, status=status.HTTP_200_OK)
        except Exception as api_exception:
            content = {"message": str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    @method_decorator(api_staff_member_required())
    def post(self, request):
        """Create a bucket

        Parameters:

            {
                "label": "label",
                "types": ["<type_version_manager_id>"]
            }

        Args:

            request: HTTP request

        Returns:

            - code: 201
              content: Created bucket
            - code: 400
              content: Validation error
            - code: 500
              content: Internal server error
        """
        try:
            # Build serializer
            bucket_serializer = BucketSerializer(data=request.data)

            # Validate data
            bucket_serializer.is_valid(True)
            # Save data
            bucket_serializer.save()

            # Return the serialized data
            return Response(bucket_serializer.data, status=status.HTTP_201_CREATED)
        except ValidationError as validation_exception:
            content = {"message": validation_exception.detail}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except Exception as api_exception:
            content = {"message": str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


class BucketDetail(APIView):
    """Retrieve, update or delete a bucket"""

    permission_classes = (IsAuthenticated,)

    def get_object(self, pk):
        """Get bucket from db

        Args:

            pk: ObjectId

        Returns:

            Bucket
        """
        try:
            return bucket_api.get_by_id(pk)
        except DoesNotExist:
            raise Http404

    def get(self, request, pk):
        """Retrieve a bucket

        Args:

            request: HTTP request
            pk: ObjectId

        Returns:

            - code: 200
              content: Bucket
            - code: 404
              content: Object was not found
            - code: 500
              content: Internal server error
        """
        try:
            # Get object
            bucket = self.get_object(pk)

            # Serialize object
            serializer = BucketSerializer(bucket)

            # Return response
            return Response(serializer.data)
        except Http404:
            content = {"message": "Bucket not found."}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as api_exception:
            content = {"message": str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    @method_decorator(api_staff_member_required())
    def delete(self, request, pk):
        """Delete a bucket

        Args:

            request: HTTP request
            pk: ObjectId

        Returns:

            - code: 204
              content: Deletion succeed
            - code: 404
              content: Object was not found
            - code: 500
              content: Internal server error
        """
        try:
            # Get object
            bucket = self.get_object(pk)

            # delete object
            bucket_api.delete(bucket)

            # Return response
            return Response(status=status.HTTP_204_NO_CONTENT)
        except Http404:
            content = {"message": "Bucket not found."}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as api_exception:
            content = {"message": str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    @method_decorator(api_staff_member_required())
    def patch(self, request, pk):
        """Update bucket

        Parameters:

            {
                "label": "<label>"
            }

        Args:

            request: HTTP request
            pk: ObjectId

        Returns:

            - code: 200
              content: Updated bucket
            - code: 400
              content: Validation error
            - code: 404
              content: Object was not found
            - code: 500
              content: Internal server error
        """
        try:
            # Get object
            bucket = self.get_object(pk)

            # Build serializer
            bucket_serializer = BucketSerializer(
                instance=bucket, data=request.data, partial=True
            )

            # Validate data
            bucket_serializer.is_valid(True)
            # Save data
            bucket_serializer.save(user=request.user)

            return Response(bucket_serializer.data, status=status.HTTP_200_OK)
        except ValidationError as validation_exception:
            content = {"message": validation_exception.detail}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except Http404:
            content = {"message": "Bucket not found."}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as api_exception:
            content = {"message": str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


class TypeVersionManagerBuckets(AbstractTemplateVersionManagerDetail):
    """Set new list of buckets for a type version manager"""

    @method_decorator(api_staff_member_required())
    def patch(self, request, pk):
        """Set new list of buckets for a type version manager

        Parameters:

            [
                {
                    "id":"<bucket_id>"
                },
                {
                    "id":"<bucket_id>"
                }
            ]

        Args:

            request: HTTP request
            pk: ObjectId

        Returns:

            - code: 200
              content: None
            - code: 400
              content: Validation error
            - code: 403
              content: Authentication error
            - code: 404
              content: Object was not found
            - code: 500
              content: Internal server error
        """
        try:
            # Get object
            type_version_manager = self.get_object(pk)

            # Serialize data
            serializer = BucketsSerializer(data=request.data, many=True)

            # Validate data
            serializer.is_valid(True)

            # Get list of unique ids
            bucket_ids = set([bucket["id"] for bucket in serializer.validated_data])

            # act
            bucket_api.update_type_buckets(type_version_manager, bucket_ids)

            return Response(status=status.HTTP_200_OK)
        except Http404:
            content = {"message": "Object not found."}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except AccessControlError as access_error:
            content = {"message": str(access_error)}
            return Response(content, status=status.HTTP_403_FORBIDDEN)
        except ValidationError as validation_exception:
            content = {"message": validation_exception.detail}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except Exception as api_exception:
            content = {"message": str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)



"""
Type API
"""
from core_composer_app.components.type.models import Type
from core_composer_app.utils.xml import check_type_core_support, COMPLEX_TYPE
from core_main_app.access_control.api import is_superuser
from core_main_app.access_control.decorators import access_control
from core_main_app.components.template import api as template_api
from core_main_app.components.template.access_control import can_write, can_read


@access_control(can_write)
def upsert(type_object, request):
    """Save or update the type.

    Args:
        type_object:
        request:

    Returns:

    """
    # Check that the type is supported by the core
    type_definition = check_type_core_support(type_object.content)
    type_object.is_complex = type_definition == COMPLEX_TYPE
    # Save type
    return template_api.upsert(type_object, request=request)


@access_control(can_read)
def get(type_id, request):
    """Get a type.

    Args:
        type_id:
        request:

    Returns:

    """
    return Type.get_by_id(type_id)


@access_control(is_superuser)
def get_all(request):
    """List all types.

    Returns:

    """
    return Type.get_all()


@access_control(is_superuser)
def get_all_complex_type(request):
    """List all complex types.

    Returns:

    """
    return Type.get_all_complex_type()


@access_control(can_write)
def delete(type_object, request):
    """Delete the type.

    Returns:

    """
    type_object.delete()
